{************************************************************************}
{*                                                                      *}
{*                         Keyboard input routines.                     *}
{*                                                                      *}
{*                          (C) J.N.Zaitseff, 1993.                     *}
{*                                                                      *}
{************************************************************************}

unit Keyb;

{ This unit provides keyboard handling routines, including the extra keys
  present on an Enhanced 101 Key keyboard. }

{$IFNDEF MSDOS}
{$IFNDEF DPMI}
    Error: MS-DOS / DPMI compilation only.
{$ENDIF}
{$ENDIF}

{ $DEFINE DEBUG}

{$A+,B-,E-,F-,N-,O-,P-,T+,V+,X+}
{$IFDEF DPMI} {$G+,F-,O-} {$ELSE} {$G-,F+,O+} {$ENDIF}
{$IFDEF DEBUG} {$D+,I+,L+,Q+,R+,S+,Y+} {$ELSE} {$D-,I-,Q-,R-,S-} {$ENDIF}

interface

type
  KeyStr   = string[2];

function EnhancedKeyboardPresent : boolean;
{ Returns TRUE if an enhanced (101/102) keyboard is present }

function Inkey : KeyStr;
{ Return a keypress. If no key has been pressed, wait until one has. }

function ScanKey : KeyStr;
{ Return a keypress, or '' if no key has been pressed. }

function Answer : char;
{ Return 'Y' or 'N'. No message is displayed. }

procedure WaitforKey;
{ Wait for a keypress }

function CapsLockIsDown         : boolean;
function CapsLockIsPressed      : boolean;

function NumLockIsDown          : boolean;
function NumLockIsPressed       : boolean;

function ScrollLockIsDown       : boolean;
function ScrollLockIsPressed    : boolean;

function AltKeyIsPressed        : boolean;
function LeftAltKeyIsPressed    : boolean;
function RightAltKeyIsPressed   : boolean;

function CtrlKeyIsPressed       : boolean;
function LeftCtrlKeyIsPressed   : boolean;
function RightCtrlKeyIsPressed  : boolean;

function ShiftKeyIsPressed      : boolean;
function LeftShiftKeyIsPressed  : boolean;
function RightShiftKeyIsPressed : boolean;

procedure ConvertWSKey (var K : KeyStr);
{ Convert simple control character presses to extended keypresses. }

implementation

var
  IsEnhanced : boolean;

function EnhancedKeyboardPresent;

var
  ShiftState : ^byte;
  Int16State : byte;

begin
  ShiftState := Ptr(Seg0040,$0017);
  EnhancedKeyboardPresent := false;
  asm
        mov     ah,12h
        int     16h
        mov     Int16State,al
  end;
  if Int16State <> ShiftState^ then
    exit;
  ShiftState^ := ShiftState^ xor $20;
  asm
        mov     ah,12h
        int     16h
        mov     Int16State,al
  end;
  EnhancedKeyboardPresent := (Int16State = ShiftState^);
  ShiftState^ := ShiftState^ xor $20
end;

function ScanKey : KeyStr; assembler;

asm
        les     bx,@Result
        mov     byte ptr es:[bx],0
        mov     ah,01h
        cmp     IsEnhanced,true
        jnz     @1
        mov     ah,11h
@1:     push    bx
        push    es
        int     16h                     { Get keyboard status }
        pop     es
        pop     bx
        jz      @end                    { Z = no key pressed. }
        mov     ah,00h
        cmp     IsEnhanced,true
        jnz     @2
        mov     ah,10h
@2:     push    bx
        push    es
        int     16h                     { Read in key }
        pop     es
        pop     bx
        cmp     ax,224                  { Check for ASCII chr 224 }
        jz      @ascii
        cmp     al,0                    { Extended key? }
        jz      @exkey
        cmp     al,224
        jz      @exkey
@ascii: mov     byte ptr es:[bx],1      { Return ASCII keypress }
        jmp     @end
@exkey: mov     al,0                    { Return extended keypress }
        mov     byte ptr es:[bx],2
@end:   inc     bx
        mov     word ptr es:[bx],ax
end;

function InKey : KeyStr;

var
  St : KeyStr;

begin
  repeat
    St := ScanKey
  until St <> '';
  InKey := St
end;

procedure WaitforKey;

var
  K : KeyStr;

begin
  K := InKey
end;

function Answer : char;

var
  K : KeyStr;

begin
  repeat
    K := InKey;
    K[1] := upcase(K[1])
  until K[1] in ['Y','N'];
  Answer := K[1]
end;

function KStat : integer; assembler;

asm
        push    bp
        cmp     IsEnhanced,true         { Check for enhanced keyboard }
        jnz     @ordkb

        mov     ah,12h                  { Read enhanced keyboard status }
        int     16h
        jmp     @end

@ordkb: mov     ah,2                    { Read ordinary keyboard status }
        int     16h
        mov     ah,0                    { "Manufacture" enhanced status }
        test    al,00000100b            { CTRL key held down? }
        jz      @1
        or      ah,00000101b
@1:     test    al,00001000b            { ALT key held down? }
        jz      @end
        or      ah,00001010b

@end:   pop     bp
end;

function CapsLockIsDown : boolean;

begin
  CapsLockIsDown := (KStat and $40) <> 0
end;

function CapsLockIsPressed : boolean;

begin
  CapsLockIsPressed := (KStat and $4000) <> 0
end;

function NumLockIsDown : boolean;

begin
  NumLockIsDown := (KStat and $20) <> 0
end;

function NumLockIsPressed : boolean;

begin
  NumLockIsPressed := (KStat and $2000) <> 0
end;

function ScrollLockIsDown : boolean;

begin
  ScrollLockIsDown := (KStat and $10) <> 0
end;

function ScrollLockIsPressed : boolean;

begin
  ScrollLockIsPressed := (KStat and $1000) <> 0
end;

function AltKeyIsPressed : boolean;

begin
  AltKeyIsPressed := (KStat and $08) <> 0
end;

function LeftAltKeyIsPressed : boolean;

begin
  LeftAltKeyIsPressed := (KStat and $0200) <> 0
end;

function RightAltKeyIsPressed : boolean;

begin
  RightAltKeyIsPressed := (KStat and $0800) <> 0
end;

function CtrlKeyIsPressed : boolean;

begin
  CtrlKeyIsPressed := (KStat and $04) <> 0
end;

function LeftCtrlKeyIsPressed : boolean;

begin
  LeftCtrlKeyIsPressed := (KStat and $0100) <> 0
end;

function RightCtrlKeyIsPressed : boolean;

begin
  RightCtrlKeyIsPressed := (KStat and $0400) <> 0
end;

function LeftShiftKeyIsPressed : boolean;

begin
  LeftShiftKeyIsPressed := (KStat and $02) <> 0
end;

function RightShiftKeyIsPressed : boolean;

begin
  RightShiftKeyIsPressed := (KStat and $01) <> 0
end;

function ShiftKeyIsPressed : boolean;

begin
  ShiftKeyIsPressed := (KStat and $03) <> 0
end;

procedure ConvertWSKey (var K : KeyStr);

begin
  if length(K) = 1 then
    case K[1] of
      ^S : K := #0#75;   { <Left Arrow> }
      ^D : K := #0#77;   { <Right Arrow> }
      ^E : K := #0#72;   { <Up Arrow> }
      ^X : K := #0#80;   { <Down Arrow> }
      ^R : K := #0#73;   { <Page Up> }
      ^C : K := #0#81;   { <Page Down> }
      ^A : K := #0#71;   { <HOME> }
      ^F : K := #0#79;   { <END> }
      ^V : K := #0#82;   { <INS> }
      ^Y : K := #0#163;  { <ALT DEL> }
      ^G : K := #0#83    { <DEL> }
    end
end;

begin
  IsEnhanced := EnhancedKeyboardPresent
end.
