{************************************************************************}
{*                                                                      *}
{*                      Mouse Interface Routines                        *}
{*                                                                      *}
{*                       (C) J.N.Zaitseff, 1993.                        *}
{*                                                                      *}
{************************************************************************}

unit Mouse;

{ This unit provides mouse interface routines. It can be used in any mode,
  both text and graphics. "InitMouse" must be the first function called. }

{$IFNDEF MSDOS}
{$IFNDEF DPMI}
    Error: MS-DOS / DPMI compilation only.
{$ENDIF}
{$ENDIF}

{ $DEFINE DEBUG}

{$A+,B-,E-,F-,N-,O-,P-,T+,V+,X+}
{$IFDEF DPMI} {$G+,F-,O-} {$ELSE} {$G-,F+,O+} {$ENDIF}
{$IFDEF DEBUG} {$D+,I+,L+,Q+,R+,S+,Y+} {$ELSE} {$D-,I-,Q-,R-,S-} {$ENDIF}

interface

procedure InitMouse;
{ Initialise the mouse driver }

function MouseInstalled : boolean;
{ Returns TRUE if the mouse driver was present when InitMouse was called }

function MouseDriverVersion : word;
{ Returns the mouse driver version, if installed }


procedure DisableMouseDriver;
{ Disable the mouse driver, if any }

procedure EnableMouseDriver;
{ Enable the mouse driver, if any }

function MouseDriverEnabled : boolean;
{ Returns TRUE if mouse driver is enabled. Driver v6.26 or higher only }

procedure ResetMouseDriver;
{ Reset the mouse driver }


procedure SaveMouseDriverState (var State : pointer);
{ Returns a pointer to the current mouse driver state. If NIL is returned,
  then the driver state could not be saved. }

procedure RestoreMouseDriverState (var State : pointer);
{ Restores the mouse driver state, and disposes of the information }


procedure ShowPointer;
{ Show the pointer, if hidden }

procedure HidePointer;
{ Hide the pointer }

procedure RestorePointer;
{ Restore the pointer to its state before HidePointer was called }

function PointerDisplayed : boolean;
{ Returns TRUE if the pointer is displayed }


procedure SetMousePosition (X, Y : integer);
{ Set the mouse position coordinate }

procedure GetMousePosition (var X, Y : integer);
{ Get the mouse position }

procedure SetMouseHorizRange (Left, Right : integer);
{ Restrict the horizontal range of travel }

procedure SetMouseVertRange (Top, Bottom : integer);
{ Restrict the vertical range of travel }

procedure GetMaxMouseRange (var X, Y : integer);
{ Get maximum virtual coordinates for mouse. Driver v7.05 or higher only }

function LeftMouseButtonDown : boolean;
{ Returns TRUE if the left mouse button is currently down }

function RightMouseButtonDown : boolean;
{ Returns TRUE if the right mouse button is currently down }

function LeftButtonPressCount (var X, Y : integer) : integer;
{ Returns the number of times the left mouse button was pressed since the
  last call to this function. (X,Y) become the coordinate of the last such
  button press }

function LeftButtonReleaseCount (var X, Y : integer) : integer;
{ Returns the number of times the left mouse button was released since the
  last call to this function. (X,Y) become the coordinate of the last such
  button press }

function RightButtonPressCount (var X, Y : integer) : integer;
{ Returns the number of times the right mouse button was pressed since the
  last call to this function. (X,Y) become the coordinate of the last such
  button press }

function RightButtonReleaseCount (var X, Y : integer) : integer;
{ Returns the number of times the right mouse button was released since the
  last call to this function. (X,Y) become the coordinate of the last such
  button press }


procedure MouseMotionCount (var X, Y : integer);
{ Returns the motion count in the horizontal and vertical directions since
  the last time this function was called }

implementation

var
  MouseInst,
  PointerOn,
  PointerWasOn : boolean;

procedure InitMouse; assembler;

asm
        mov     PointerOn,false
        mov     PointerWasOn,true
        mov     ax,0
        int     33h
        mov     MouseInst,false
        cmp     ax,0
        jz      @1
        mov     MouseInst,true
@1:
end;

function MouseInstalled : boolean;

begin
  MouseInstalled := MouseInst
end;

function MouseDriverVersion : word; assembler;

asm
        cmp     MouseInst,true
        jz      @1

        xor     ax,ax
        jmp     @2

@1:     mov     ax,24h          { Get mouse driver version }
        int     33h
        mov     ax,bx
@2:
end;

procedure DisableMouseDriver; assembler;

asm
        mov     ax,1Fh
        push    es
        int     33h
        pop     es
end;

procedure EnableMouseDriver; assembler;

asm
        mov     ax,20h
        int     33h
end;

function MouseDriverEnabled : boolean; assembler;

asm
        mov     al,MouseInst
        cmp     al,true
        jnz     @1

        mov     ax,26h
        int     33h

        mov     al,true
        cmp     bl,0
        jz      @1
        mov     al,false

@1:
end;

procedure ResetMouseDriver; assembler;

asm
        mov     PointerOn,false;
        mov     PointerWasOn,true;
        mov     ax,21h
        int     33h

        mov     bl,true
        cmp     ax,21h
        jnz     @1
        mov     bl,false

@1:     mov     MouseInst,bl
end;

const
  WordSize = sizeof(word);

procedure SaveMouseDriverState (var State : pointer);

var
  BufSize : word;

begin
  if not MouseInst then
    State := nil
  else
    begin
      asm
        mov     ax,15h
        int     33h
        mov     BufSize,bx
      end;
      GetMem(State,BufSize+WordSize);
      asm
        les     bx,State[WordSize]
        mov     ax,BufSize
        mov     es:[bx],ax
        mov     dx,bx
        mov     ax,16h
        int     33h
      end
    end
end;

procedure RestoreMouseDriverState (var State : pointer);

var
  BufSize : word;

begin
  if (State <> nil) then
    begin
      asm
        les     bx,State[WordSize]
        mov     ax,es:[bx]
        mov     BufSize,ax
        mov     ax,17h
        int     33h
      end;
      FreeMem(State,BufSize+WordSize);
      State := nil
    end
end;

procedure ShowPointer; assembler;

asm
        cmp     PointerOn,true
        jz      @1

        mov     ax,1
        int     33h
        mov     PointerOn,true

@1:
end;

procedure HidePointer; assembler;

asm
        mov     al,PointerOn
        mov     PointerWasOn,al
        cmp     al,true
        jnz     @1

        mov     ax,2
        int     33h
        mov     PointerOn,false
@1:
end;

procedure RestorePointer;

begin
  if PointerWasOn then
    ShowPointer
end;

function PointerDisplayed : boolean;

begin
  PointerDisplayed := PointerOn
end;

procedure SetMousePosition (X, Y : integer); assembler;

asm
        mov     ax,4
        mov     cx,X
        mov     dx,Y
        int     33h
end;

procedure GetMousePosition (var X, Y : integer); assembler;

asm
        cmp     MouseInst,true
        jz      @1

        xor     ax,ax
        mov     cx,ax
        mov     dx,ax
        jmp     @2

@1:     mov     ax,3
        int     33h

@2:     les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
end;

procedure SetMouseHorizRange (Left, Right : integer); assembler;

asm
        mov     ax,7
        mov     cx,Left
        mov     dx,Right
        int     33h
end;

procedure SetMouseVertRange (Top, Bottom : integer); assembler;

asm
        mov     ax,8
        mov     cx,Top
        mov     dx,Bottom
        int     33h
end;

procedure GetMaxMouseRange (var X, Y : integer); assembler;

asm
        xor     ax,ax
        mov     cx,ax
        mov     dx,ax

        cmp     MouseInst,true
        jnz     @1

        mov     ax,37h
        int     33h

@1:     les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
end;

function LeftMouseButtonDown : boolean; assembler;

asm
        mov     al,false
        cmp     MouseInst,true
        jnz     @end

        mov     ax,3
        int     33h
        and     bl,00000001b
        mov     al,bl
@end:
end;

function RightMouseButtonDown : boolean; assembler;

asm
        mov     al,false
        cmp     MouseInst,true
        jnz     @end

        mov     ax,3
        int     33h
        mov     al,false
        test    bl,00000010b
        jz      @end
        mov     al,true
@end:
end;

function LeftButtonPressCount (var X, Y : integer) : integer; assembler;

asm
        mov     ax,0
        cmp     MouseInst,true
        jnz     @end

        mov     ax,5
        mov     bx,0
        int     33h
        mov     ax,bx
        les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
@end:
end;

function LeftButtonReleaseCount (var X, Y : integer) : integer; assembler;

asm
        mov     ax,0
        cmp     MouseInst,true
        jnz     @end

        mov     ax,6
        mov     bx,0
        int     33h
        mov     ax,bx
        les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
@end:
end;

function RightButtonPressCount (var X, Y : integer) : integer; assembler;

asm
        mov     ax,0
        cmp     MouseInst,true
        jnz     @end

        mov     ax,5
        mov     bx,1
        int     33h
        mov     ax,bx
        les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
@end:
end;

function RightButtonReleaseCount (var X, Y : integer) : integer; assembler;

asm
        mov     ax,0
        cmp     MouseInst,true
        jnz     @end

        mov     ax,6
        mov     bx,1
        int     33h
        mov     ax,bx
        les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
@end:
end;

procedure MouseMotionCount (var X, Y : integer); assembler;

asm
        cmp     MouseInst,true
        jz      @1

        xor     ax,ax
        mov     cx,ax
        mov     dx,ax
        jmp     @2

@1:     mov     ax,11
        int     33h

@2:     les     bx,X
        mov     es:[bx],cx
        les     bx,Y
        mov     es:[bx],dx
end;


begin
  MouseInst := false;
  PointerOn := false;
  PointerWasOn := true
end.
