{**************************************************************************}
{*                                                                        *}
{*                           Text Window Routines                         *}
{*                                                                        *}
{*                          (C) J.N.Zaitseff, 1994.                       *}
{*                                                                        *}
{**************************************************************************}

{  This unit creates and manipulates pop-up text windows. These windows may
   be either in memory or stored on disk.

   InitTxWindows must be called before any windows are opened. Subsequent
   calls may be made to DrawDefaultBackground, then to OpenWindow,
   CloseWindow, and so on.

   By default, windows are saved to memory. This can be changed BEFORE
   InitTxWindows is called by calling SetWindowStorageToMemory or
   SetWindowStorageToDisk. Window files are stored in the directory specified
   by the DOS "TEMP" variable, and use the DOS "generate-unique-name" system
   call. Each window requires 4016 bytes of memory (if storing to memory), or
   4000 bytes disk / 16 bytes memory (if storing to disk). An absolute
   maximum of 46656 windows may be opened.

   At the moment, this unit can only handle screens of size 80 x 25.
}

unit TxWindows;

{$IFNDEF MSDOS}
{$IFNDEF DPMI}
    Error: MS-DOS / DPMI compilation only.
{$ENDIF}
{$ENDIF}

{ $DEFINE DEBUG}

{$A+,B-,E-,N-,P-,T+,V+,X+}
{$IFDEF DPMI} {$G+,F-,O-} {$ELSE} {$G-,F+,O+} {$ENDIF}
{$IFDEF DEBUG} {$D+,I+,L+,Q+,R+,S+,Y+} {$ELSE} {$D-,I-,Q-,R-,S-} {$ENDIF}

interface

const
  NoError           = 0;   { No error occurred }
  TxWindowNotInited = 1;   { InitTxWindow has not been called yet }
  TxWindowNotOpen   = 2;   { No open window - could not close }
  TxWindowNotFound  = 3;   { Could not close window - file missing on disk }
  TxWindowNotSaved  = 4;   { Could not save window - out of memory/disk space }
  TxWindowCoordErr  = 5;   { Coordinate out of range error }
  TxWindowStarted   = 6;   { TxWindows has already been in use }

  NoBorder          = 0;   { Border types }
  SingleLineBorder  = 1;
  DoubleLineBorder  = 2;
  ThickBorder       = 3;
  LastBorderOption  = ThickBorder;


procedure SetWindowStorageToDisk;
{ Set the storage of text windows to disk. Each window takes 4000 bytes of
  disk space and 16 bytes of memory (if stored to disk), or 4016 bytes of
  memory. For disk, windows are saved in the directory specified by the DOS
  "TEMP" variable (if set), and uses a unique filename. This procedure must
  be called BEFORE InitTxWindows. }

procedure SetWindowStorageToMemory;
{ Set the storage of text windows to memory. Memory is allocated dynamically
  for each window. This procedure must be called BEFORE InitTxWindows. }

procedure InitTxWindows;
{ Initialise internal routines. The screen is NOT affected, although the text
  viewport is. }

procedure DrawDefaultBackground;
{ Clear the screen - the default background is black. Note that this must be
  called before ANY windows are opened, or not at all. }

procedure OpenWindow (X,Y, Width, Height, FCol, BCol, BdCol, BdType : integer);
{ Open a window at (X,Y), having Width, Height. FCol = Foreground colour,
  BCol = Background colour of window, BdCol = Border colour, BdType = border
  type (NoBorder, SingleLineBorder, DoubleLineBorder, ThickBorder).
  The border is included in the width and height of the window. Unless
  NoBorder is specified, the actual text area will be (Width-4) x (Height-2).
  The viewport is changed to be from (1,1) to (Width,Height) if NoBorder
  was specified, or (1,1) to (Width-4,Height-2) otherwise. }

procedure CloseWindow;
{ Close last window opened, restoring what was underneath }

procedure CloseAllWindows;
{ Close all windows opened }

procedure DeleteAllWindowData;
{ Call this procedure just before the program ends to delete all window data
  from memory or the disk. Note that the screen is not refreshed. This is a
  quicker way than CloseAllWindows. }

procedure SetWindowViewPort;
{ Restore the window viewport so that the coordinates are (1,1) to
  (Width,Height) for a window with no border, or (Width-4,Height-2) otherwise.
  The cursor is placed at (1,1). The window itself is not affected. }

procedure SetDefaultColours;
{ Restores the foreground and background colours for the window. Note that
  the window itself is not affected }

procedure CentreText (const St : string; Line : integer);
{ Centre the string "St" on line "Line" in the window using the current
  colours. }

function NumberWindows : integer;
{ Returns the number of windows currently open. }

function WinWidth : byte;
{ Returns the width of the text area of the current window. }

function WinHeight : byte;
{ Returns the height of the text area of the current window. }

function InWindow (X, Y : integer) : boolean;
{ Returns TRUE if the global coordinate (X,Y) is in the last opened window. }

function WindowError : byte;
{ Returns the last error code }


implementation

uses
  Crt, Dos, Mouse;

type
  TxtElem   = record                    { Text screen element: }
                Ch : char;              {   character          }
                At : byte               {   attribute          }
              end;
  ScnArray  = array [1..25,1..80] of TxtElem;
  WinRecPtr = ^WindowRec;
  WindowRec = record
                WX1, WY1,             { Starting X,Y coords of text area  }
                WX2, WY2,             { Ending X,Y coords of text area    }
                FC, BC,               { Foreground, Background colour     }
                CX, CY  : byte;       { Previous position of cursor       }
                Scn     : ^ScnArray;  { Pointer to screen data            }
                Prev    : WinRecPtr   { Pointer to previous window        }
              end;

const
  BdArray : array [SingleLineBorder..ThickBorder,1..6] of char
              = ('Ŀ','ͻȼ','');

var
  Screen         : ^ScnArray;
  CurrentWindow,
  LastWindow     : WinRecPtr;
  WindowPath     : PathStr;
  UseDisk        : boolean;
  NumWindows     : integer;
  WindowErr      : byte;
  Inited         : boolean;


procedure SetWindowStorageToDisk;

begin
  if Inited then
    WindowErr := TxWindowStarted
  else
    UseDisk := true
end;

procedure SetWindowStorageToMemory;

begin
  if Inited then
    WindowErr := TxWindowStarted
  else
    UseDisk := false
end;

procedure InitTxWindows;

begin
  if Inited then
    WindowErr := TxWindowStarted
  else
    begin
      Inited := true;
      WindowErr := NoError;
      Window(1,1,80,25)
    end
end;

procedure DrawDefaultBackground;

begin
  WindowErr := NoError;
  HidePointer;
  Window(1,1,80,25);
  TextColor(LightGray);
  TextBackground(Black);
  ClrScr;
  RestorePointer
end;

type
  str3 = string[3];

function encode (N : integer) : str3;

const
  Tbl : array[0..35] of char = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';

var
  T : str3;

begin
  T[0] := #3;
  T[1] := Tbl[N mod 36];
  T[2] := Tbl[(N div 36) mod 36];
  T[3] := Tbl[(N div 1296) mod 36];
  encode := T
end;

function SaveWindow (var W : WinRecPtr; WindowNum : integer) : boolean;

var
  F  : file;
  OK : boolean;

begin
  if UseDisk then
    begin
      W^.Scn := nil;
      assign(F,WindowPath+encode(WindowNum)+'.$$$');
{$I-}
      rewrite(F,1);
      OK := (IOResult = 0);
      if OK then
        begin
          HidePointer;
          blockwrite(F,Screen^,sizeof(Screen^));
          RestorePointer;
          OK := (IOResult = 0);
          if OK then
            begin
              close(F);
              OK := (IOResult = 0)
            end
        end;
{$IFDEF DEBUG} {$I+} {$ENDIF}
      SaveWindow := OK
    end
  else
    if MaxAvail < sizeof(ScnArray) then
      SaveWindow := false
    else
      with W^ do
        begin
          new(Scn);
          HidePointer;
          Scn^ := Screen^;
          RestorePointer;
          SaveWindow := true
        end
end;

function LoadWindow (W : WinRecPtr; WindowNum : integer) : boolean;

var
  F  : file;
  OK : boolean;

begin
  if UseDisk then
    begin
      assign(F,WindowPath+encode(WindowNum)+'.$$$');
{$I-}
      reset(F,1);
      OK := (IOResult = 0);
      if OK then
        begin
          HidePointer;
          blockread(F,Screen^,sizeof(Screen^));
          RestorePointer;
          close(F);
          OK := (IOResult = 0);
          if OK then
            begin
              erase(F);
              OK := (IOResult = 0)
            end
        end;
{$IFDEF DEBUG} {$I+} {$ENDIF}
      LoadWindow := OK
    end
  else
    with W^ do
      begin
        HidePointer;
        Screen^ := Scn^;
        RestorePointer;
        dispose(Scn);
        LoadWindow := true;
      end
end;

procedure DrawWindow (X,Y, Width, Height, FCol, BCol, BdCol, BdType : integer);

var
  I : integer;

begin
  HidePointer;
  Window(X,Y,X+Width-1,Y+Height-1);
  TextColor(BdCol);
  TextBackground(BCol);
  ClrScr;
  Window(1,1,80,25);
  if BdType <> NoBorder then
    begin
      GotoXY(X,Y);
      write(BdArray[BdType,1]);
      for I := 2 to Width-1 do
        write(BdArray[BdType,2]);
      write(BdArray[BdType,3]);
      for I := 1 to Height-2 do
        begin
          GotoXY(X,Y+I);
          write(BdArray[BdType,4]);
          GotoXY(X+Width-1,Y+I);
          write(BdArray[BdType,4])
        end;
      GotoXY(X,Y+Height-1);
      write(BdArray[BdType,5]);
      for I := 2 to Width-1 do
        write(BdArray[BdType,2]);
      Screen^[Y+Height-1,X+Width-1].Ch := BdArray[BdType,6]
    { write(BdArray[BdType,6]) }
    end;
  RestorePointer
end;

procedure OpenWindow (X,Y, Width, Height, FCol, BCol, BdCol, BdType : integer);

var
  NewWindow : WinRecPtr;

begin
  if (X < 1) or (X > 80) or (Y < 1) or (Y > 25) or (Width < 1) or
     (X + Width > 81) or (Height < 1) or (Y + Height > 26) then
    begin
      WindowErr := TxWindowCoordErr;
      exit
    end;

  new(NewWindow);
  with NewWindow^ do
    begin
      FC := FCol;
      BC := BCol;
      CX := WhereX;
      CY := WhereY;
      if BdType = NoBorder then
        begin
          WX1 := X;
          WY1 := Y;
          WX2 := X + Width - 1;
          WY2 := Y + Height - 1
        end
      else
        begin
          WX1 := X + 2;
          WY1 := Y + 1;
          WX2 := X + Width - 3;
          WY2 := Y + Height - 2
        end
    end;

  if not SaveWindow(NewWindow, NumWindows) then
    begin
      dispose(NewWindow);
      WindowErr := TxWindowNotSaved;
      exit
    end;

  if NumWindows = 0 then
    begin
      CurrentWindow := NewWindow;
      CurrentWindow^.Prev := nil
    end
  else
    begin
      LastWindow := CurrentWindow;
      CurrentWindow := NewWindow;
      CurrentWindow^.Prev := LastWindow
    end;

  inc(NumWindows);

  DrawWindow(X,Y, Width,Height, FCol,BCol,BdCol, BdType);
  with CurrentWindow^ do
    Window(WX1,WY1, WX2, WY2);
  TextColor(FCol);

  WindowErr := NoError
end;

procedure CloseWindow;

var
  OldWindow : WinRecPtr;

begin
  if NumWindows = 0 then
    begin
      WindowErr := TxWindowNotOpen;
      exit
    end;

  if not LoadWindow(CurrentWindow,NumWindows - 1) then
    begin
      WindowErr := TxWindowNotFound;
      exit
    end;

  OldWindow := CurrentWindow;

  if NumWindows = 1 then
    begin
      Window(1,1,80,25);
      with OldWindow^ do
        GotoXY(CX,CY);
      TextColor(LightGray);
      TextBackground(Black);
      dispose(OldWindow)
    end
  else
    begin
      CurrentWindow := LastWindow;
      LastWindow := LastWindow^.Prev;
      with CurrentWindow^ do
        begin
          Window(WX1,WY1, WX2,WY2);
          TextColor(FC);
          TextBackground(BC)
        end;
      with OldWindow^ do
        GotoXY(CX,CY);
      dispose(OldWindow)
    end;

  dec(NumWindows);

  WindowErr := NoError
end;

procedure CloseAllWindows;

var
  I, J : integer;

begin
  WindowErr := TxWindowNotOpen;
  J := NumWindows;
  for I := 1 to J do
    CloseWindow
end;

procedure DeleteAllWindowData;

var
  F    : file;
  I    : integer;
  p, q : WinRecPtr;

begin
  p := CurrentWindow;
  WindowErr := NoError;
  for I := 1 to NumWindows do
    begin
      if UseDisk then
        begin
          assign(F,WindowPath+encode(I-1)+'.$$$');
{$I-}
          erase(F);
{$IFDEF DEBUG} {$I+} {$ENDIF}
          if IOResult <> 0 then
            WindowErr := TxWindowNotFound
        end
      else
        dispose(p^.Scn);
      q := p^.Prev;
      dispose(p);
      p := q
    end;
  CurrentWindow := nil;
  LastWindow := nil;
  NumWindows := 0;
end;

procedure SetWindowViewPort;

begin
  if NumWindows = 0 then
    WindowErr := TxWindowNotOpen
  else
    begin
      with CurrentWindow^ do
        Window(WX1,WY1, WX2,WY2);
      WindowErr := NoError
    end
end;

procedure SetDefaultColours;

begin
  if NumWindows = 0 then
    WindowErr := TxWindowNotOpen
  else
    with CurrentWindow^ do
      begin
        TextColor(FC);
        TextBackground(BC);
        WindowErr := NoError
      end
end;

procedure CentreText (const St : string; Line : integer);

begin
  if NumWindows = 0 then
    WindowErr := TxWindowNotOpen
  else
    with CurrentWindow^ do
      begin
        GotoXY((WX2 - WX1 - length(St) + 3) div 2,Line);
        write(St);
        GotoXY(1,Line+1);
        WindowErr := NoError
      end
end;

function NumberWindows : integer;

begin
  NumberWindows := NumWindows
end;

function WinWidth : byte;

begin
  if NumWindows = 0 then
    WindowErr := TxWindowNotOpen
  else
    begin
      WindowErr := NoError;
      with CurrentWindow^ do
        WinWidth := WX2 - WX1 + 1
    end
end;

function WinHeight : byte;

begin
  if NumWindows = 0 then
    WindowErr := TxWindowNotOpen
  else
    begin
      WindowErr := NoError;
      with CurrentWindow^ do
        WinHeight := WY2 - WY1 + 1
    end
end;

function InWindow (X, Y : integer) : boolean;

begin
  if NumWindows = 0 then
    WindowErr := TxWindowNotOpen
  else
    begin
      WindowErr := NoError;
      with CurrentWindow^ do
        InWindow := (X >= WX1) and (X <= WX2) and (Y >= WY1) and (Y <= WY2)
    end
end;

function WindowError : byte;

begin
  WindowError := WindowErr
end;

function VidSeg : word;

begin
  if mem[Seg0040:$0049] = 7 then
    VidSeg := SegB000
  else
    VidSeg := SegB800
end;

function TempFileName : PathStr;

var
  T    : PathStr;
  S    : string[4];
  I, J : integer;
  N    : longint;

begin
  randomize;
  T := GetEnv('TEMP');
  if (T <> '') and (T[length(T)] <> '\') then
    T := T + '\';
  J := 0;
  N := RandSeed;
  for I := 1 to 8 do
    begin
      J := J + N and $F;
      N := N shr 4
    end;
  str(J:3,S);
  for I := 1 to 3 do
    if S[I] = ' ' then
      S[I] := '0';
  T := T + '#' + S;
  TempFileName := T
end;

begin { Initialisation }
{$IFNDEF DEBUG}
  CheckBreak := false;
{$ENDIF}
  Screen := Ptr(VidSeg,0);
  CurrentWindow := nil;
  LastWindow := nil;
  WindowPath := TempFileName;
  UseDisk := false;
  NumWindows := 0;
  WindowErr := NoError;
  Inited := false
end.
